! TXSUM -- Print/Check/Update Checksums
! Nils M Holm, 2001,2002,2019
! T3X/0 version, 2024
! In the public domain / under the 0BSD license

use t3x: t;
use string: str;
use strfmt: fmt;
use strparse: parser;
use io;

const	BUFLEN		= 1024;
const	BLOCKSIZE	= 512;
const	MAXPATHL	= 256;

struct	MODES = M_MAKE, M_CHECK, M_UPDATE, M_DIGEST;

var	Mode;
var	Compat;

msg(m, n) do var b::3;
	t.write(T3X.SYSERR, m, str.length(m));
	if (n) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, n, str.length(n));
	end
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
	return -1;
end

mksum(file, sum, sizes) do
	var	fd, buf::BLOCKSIZE;
	var	k, i, c, o, j, nb, kl, kh, ffff;

	fd := file-> t.open(file, T3X.OREAD): T3X.SYSIN;
	if (fd < 0) return msg("TXSUM: could not open", file);
	for (i=0, 8) sum[i] := 0;
	nb := 0;
	kl := 0;
	kh := 0;
	ffff := 0xfff << 4 | 0xf;
	while (1) do
		k := t.read(fd, buf, BLOCKSIZE);
		if (k < 0) do
			msg("TXSUM: file read error", file);
			t.close(fd);
			return -1;
		end
		if (k < BLOCKSIZE) t.memfill(@buf::k, 0, BLOCKSIZE-k);
		for (i=0, BLOCKSIZE) do
			c := sum[0] & 1;
			for (j=0, 7)
				sum[j] := sum[j] >> 1 |
					(sum[j+1] & 1 -> 0x80 : 0);
			sum[7] := sum[7] >> 1 | (c -> 0x80 : 0);
			sum[0] := sum[0] + buf::i;
			for (j=1, 8) do
				if (sum[j-1] > 0xFF) do
					sum[j] := sum[j] + 1;
					sum[j-1] := sum[j-1] & 0xFF;
				end
			end
			sum[7] := sum[7] & 0xFF;
		end
		nb := nb+1;
		if (kl + k & ffff .< kl) kh := kh + 1;
		kl := kl + k & ffff;
		if (k < BLOCKSIZE) leave;
	end
	if (file) t.close(fd);
	ie (Compat) do
		sizes[0] := nb;
		sizes[1] := k;
	end
	else do
		sizes[0] := kh;
		sizes[1] := kl;
	end
end

check_sums() do
	var	line::MAXPATHL*2, name::MAXPATHL;
	var	sum[8], osum[8], sizes[2], osizes[2];
	var	k;

	while (1) do
		k := io.readln(line, MAXPATHL);
		if (k < 1) leave;
		k := parser.parse(line, "%2X%2X%2X%2X%2X%2X%2X%2X %4X %4X %S",
			[(@osum[7], @osum[6], @osum[5], @osum[4],
			@osum[3], @osum[2], @osum[1], @osum[0],
			@osizes[0], @osizes[1], name)]);
		if (k \= 11) do
			msg("TXSUM: corrupt checksum entry", 0);
			loop;
		end
		if (mksum(name, sum, sizes) < 0) loop;
		if (sizes[0] \= osizes[0] \/ sizes[1] \= osizes[1]) do
			msg("TXSUM: file size mismatch", name);
			loop;
		end
		if (t.memcomp(sum, osum, 8*t.bpw()))
			msg("TXSUM: invalid checksum", name);
	end
end

make_sums() do var file::MAXPATHL, buf::MAXPATHL*2, k, sum[8], sizes[2];
	while (1) do
		k := io.readln(file, MAXPATHL);
		if (k < 1) leave;
		mksum(file, sum, sizes);
		io.writeln(fmt.nformat(MAXPATHL*2, buf,
		"%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X %4:0UX %4:0UX %S",
			[(sum[7],sum[6],sum[5],sum[4],
			 sum[3],sum[2],sum[1],sum[0],
			 sizes[0], sizes[1], file)]));
	end
end

update_sums() do
	var	line::MAXPATHL*2, name::MAXPATHL;
	var	sum[8], osum[8], sizes[2], osizes[2];
	var	k;

	while (1) do
		k := io.readln(line, MAXPATHL);
		if (k < 1) leave;
		k := parser.parse(line, "%2X%2X%2X%2X%2X%2X%2X%2X %4X %4X %S",
			[(@osum[7], @osum[6], @osum[5], @osum[4],
			@osum[3], @osum[2], @osum[1], @osum[0],
			@osizes[0], @osizes[1], name)]);
		if (k \= 11) do
			msg("TXSUM: corrupt checksum entry", 0);
			loop;
		end
		if (mksum(name, sum, sizes) < 0) loop;
		ie (sizes[0] \= osizes[0] \/ sizes[1] \= osizes[1])
			msg("TXSUM: updating", name);
		else if (t.memcomp(sum, osum, 8*t.bpw()))
			msg("TXSUM: updating", name);
		io.writeln(fmt.nformat(MAXPATHL*2, line,
		"%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X %4:0UX %4:0UX %S",
			[(sum[7], sum[6], sum[5], sum[4],
			  sum[3], sum[2], sum[1], sum[0],
			 sizes[0], sizes[1], name)]));
	end
end

make_digest() do var buf::MAXPATHL*2, sum[8], sizes[2];
	mksum(0, sum, sizes);
	io.writeln(fmt.nformat(MAXPATHL*2, buf,
	"%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X%2:0X",
		[(sum[7], sum[6], sum[5], sum[4],
		  sum[3], sum[2], sum[1], sum[0])]));
end

usage() do
	msg("Usage: txsum [-bcdmu]", 0);
	msg("-b - compatibility mode (report sizes in blocks)", 0);
	msg("-c - check checksums for files listed in sysin", 0);
	msg("-d - create digest for data in sysin", 0);
	msg("-m - make checksums for files listed in sysin", 0);
	msg("-u - update checksums of files listed in sysin", 0);
	halt 1;
end

do var abuf::128, p, i;
	Mode := M_CHECK;
	Compat := 0;
	p := 1;
	while (1) do
		if (t.getarg(p, abuf, 128) < 1) leave;
		if (abuf::0 \= '-') leave;
		i := 1;
		while (abuf::i) do
			     ie (abuf::i = 'c') Mode := M_CHECK;
			else ie (abuf::i = 'd') Mode := M_DIGEST;
			else ie (abuf::i = 'm') Mode := M_MAKE;
			else ie (abuf::i = 'u') Mode := M_UPDATE;
			else ie (abuf::i = 'b') Compat := %1;
			else usage();
			i := i+1;
		end
		p := p+1;
	end
	ie (Mode = M_CHECK) check_sums();
	else ie (Mode = M_MAKE) make_sums();
	else ie (Mode = M_UPDATE) update_sums();
	else if (Mode = M_DIGEST) make_digest();
end
